//
//  Implementation of the DSregionFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <sstream>
#include <cstring>
#include <cmath>
#include <cstdio>

#include <Dal.h>
#include <Region.h>
#include <ASCregion.h>
#include "DataFilter.h"
#include <DataSubspace.h>
#include <errstr.h>


DSregionFilter::DSregionFilter(DScomp *parent,
			       const std::string &name, GeoRegion region,
			       DSfilterComp &filtercomp) :
  DSfilter(parent, RegionT, name, "", filtercomp),  _region(region){
  for (unsigned n=0; n<region.size(); ++n){
    _region[n].clear();
    for (unsigned m=0; m<region[n].size(); ++m){
      switch( region[n][m]->type() ){
      case ShapeBasics::LINE:
	_region[n].push_back( new FLine(*dynamic_cast<FLine *>(region[n][m])) );
	break;
      case ShapeBasics::CIRCLE:
	_region[n].push_back( new FCircle(*dynamic_cast<FCircle *>(region[n][m])) );
	break;
      case ShapeBasics::SECTOR:
	_region[n].push_back( new FSector(*dynamic_cast<FSector *>(region[n][m])) );
	break;
      case ShapeBasics::ANNULUS:
	_region[n].push_back( new FAnnulus(*dynamic_cast<FAnnulus *>(region[n][m])) );
	break;
      case ShapeBasics::ELLIPSE:
	_region[n].push_back( new FEllipse(*dynamic_cast<FEllipse *>(region[n][m])) );
	break;
      case ShapeBasics::ELLIPTANNULUS:
	_region[n].push_back( new FElliptAnnulus(*dynamic_cast<FElliptAnnulus *>(region[n][m])) );
	break;
      case ShapeBasics::ROTRHOMBUS:
	_region[n].push_back( new FRotRhombus(*dynamic_cast<FRotRhombus *>(region[n][m])) );
	break;
      case ShapeBasics::RHOMBUS:
	_region[n].push_back( new FRhombus(*dynamic_cast<FRhombus *>(region[n][m])) );
	break;
      case ShapeBasics::ROTBOX:
	_region[n].push_back( new FRotBox(*dynamic_cast<FRotBox *>(region[n][m])) );
	break;
      case ShapeBasics::BOX:
	_region[n].push_back( new FBox(*dynamic_cast<FBox *>(region[n][m])) );
	break;
      case ShapeBasics::ROTRECTANGLE:
	_region[n].push_back( new FRotRectangle(*dynamic_cast<FRotRectangle *>(region[n][m])) );
	break;
      case ShapeBasics::RECTANGLE:
	_region[n].push_back( new FRectangle(*dynamic_cast<FRectangle *>(region[n][m])) );
	break;
      case ShapeBasics::POLYGON:
	_region[n].push_back( new FPolygon(*dynamic_cast<FPolygon *>(region[n][m])) );
	break;
      default:
	_region[n].push_back( new FShape(*region[n][m]) );
      }
    }
  }
}


DSregionFilter::DSregionFilter(DScomp *parent,
			       const std::string &name, GeoRegion region) :
  DSfilter(parent, RegionT, name), _region(region) {
  for (unsigned n=0; n<region.size(); ++n){
    _region[n].clear();
    for (unsigned m=0; m<region[n].size(); ++m){
      if( dynamic_cast<FLine *>(region[n][m]) )
	_region[n].push_back( new FLine(*dynamic_cast<FLine *>(region[n][m])) );
      else if( dynamic_cast<FCircle *>(region[n][m]) )
	_region[n].push_back( new FCircle(*dynamic_cast<FCircle *>(region[n][m])) );
      else if( dynamic_cast<FSector *>(region[n][m]) )
	_region[n].push_back( new FSector(*dynamic_cast<FSector *>(region[n][m])) );
      else if( dynamic_cast<FAnnulus *>(region[n][m]) )
	_region[n].push_back( new FAnnulus(*dynamic_cast<FAnnulus *>(region[n][m])) );
      else if( dynamic_cast<FEllipse *>(region[n][m]) )
	_region[n].push_back( new FEllipse(*dynamic_cast<FEllipse *>(region[n][m])) );
      else if( dynamic_cast<FElliptAnnulus *>(region[n][m]) )
	_region[n].push_back( new FElliptAnnulus(*dynamic_cast<FElliptAnnulus *>(region[n][m])) );
      else if( dynamic_cast<FRotRhombus *>(region[n][m]) )
	_region[n].push_back( new FRotRhombus(*dynamic_cast<FRotRhombus *>(region[n][m])) );
      else if( dynamic_cast<FRhombus *>(region[n][m]) )
	_region[n].push_back( new FRhombus(*dynamic_cast<FRhombus *>(region[n][m])) );
      else if( dynamic_cast<FRotBox *>(region[n][m]) )
	_region[n].push_back( new FRotBox(*dynamic_cast<FRotBox *>(region[n][m])) );
      else if( dynamic_cast<FBox *>(region[n][m]) )
	_region[n].push_back( new FBox(*dynamic_cast<FBox *>(region[n][m])) );
      else if( dynamic_cast<FRotRectangle *>(region[n][m]) )
	_region[n].push_back( new FRotRectangle(*dynamic_cast<FRotRectangle *>(region[n][m])) );
      else if( dynamic_cast<FRectangle *>(region[n][m]) )
	_region[n].push_back( new FRectangle(*dynamic_cast<FRectangle *>(region[n][m])) );
      else if( dynamic_cast<FPolygon *>(region[n][m]) )
	_region[n].push_back( new FPolygon(*dynamic_cast<FPolygon *>(region[n][m])) );
      else
	_region[n].push_back( new FShape(*region[n][m]) );
    }
  }
}


GeoRegion DSregionFilter::region() const{
  return _region;
}


void DSregionFilter::write(unsigned compNumber, unsigned filterNumber, 
			   Block *toBlock, bool dummy) {

  // A keyword region specification is pretty limited, so instead
  // write out region filters as references to FITS region blocks.

  std::stringstream valueName0, refName, refNum;
  std::string ref;
  
  refNum << "REG";
  refNum.width(5);
  refNum.fill('0');
  refNum << compNumber*100 + filterNumber;
  ref = refNum.str();

  valueName0 << "DSVAL" << filterNumber;

  if( !toBlock->hasAttribute(valueName0.str()) ||
      toBlock->attribute(valueName0.str())->asString() != "TABLE" )
    *toBlock->addAttribute(valueName0.str(), comment+"value") = "TABLE";

  DSfilterComp filterComps = comps();
  std::stringstream typeName;
  typeName << "DSTYP" << filterNumber;
    
  *toBlock->attribute(typeName.str()) = toBlock->attribute(typeName.str())->asString() + 
    "(" + filterComps[0] + "," + filterComps[1] + ")";

  std::stringstream refValue;
  if( dummy )
    refValue << ":";
  else{
    std::string blockSpec="NoBlockSpec";
    _getRefBlockName(compNumber, filterNumber, toBlock, blockSpec, ref) ;
    refValue << ":" << ref;
  }

  std::stringstream refName0;
  refName0 << "DSREF" << filterNumber;

  if( compNumber==1 )
    *toBlock->addAttribute(refName0.str(), comment+"reference")
      = refValue.str();
  else if( ! toBlock->hasAttribute(refName0.str()) ||
	   toBlock->attribute(refName0.str())->asString() != refValue.str() ){
    std::stringstream refName;
    refName << compNumber << "DSREF" << filterNumber;
    *toBlock->addAttribute(refName.str(), comment+"reference") 
      = refValue.str();
  }

  // Need to copy the region to the output data set.
  // First check for a name clash.

  if( !dummy ){
    Table *ascTable = toBlock->parent()->addTable(ref, _region.elements());
    DSfilterComp regionComps = comps();

    OutASCregionTable ascRegion(ascTable, regionComps[0], regionComps[1]);
    ascRegion << _region;
  }
}



std::string DSregionFilter::selectExpression() {
  std::stringstream expression;

  DSfilterComp filtComps = comps();

  expression << "(";
  for (unsigned comp=0; comp<_region.size(); ++comp){
    expression << "(";

    for (unsigned element=0; element<_region[comp].size(); ++element) {
      if(! _region[comp][element]->included() ) expression << "!";

      std::vector<float> X = _region[comp][element]->pointsX();
      std::vector<float> Y = _region[comp][element]->pointsY();
      std::vector<float> R = _region[comp][element]->radii();
      std::vector<double> angles = _region[comp][element]->rotAngles();

      switch(_region[comp][element]->type()){
      case ShapeBasics::POINT         : {
	expression << "point(";
	break;
      }
      case ShapeBasics::CIRCLE        : {
	expression << "circle(";
	break;
      }
      case ShapeBasics::ANNULUS       : {
	expression << "ring(";
	break;
      }
      case ShapeBasics::ELLIPSE       : {
	expression << "ellipse(";
	break;
      }
      case ShapeBasics::ELLIPTANNULUS : {
	expression << "elliptring(";
	break;
      }
      case ShapeBasics::SECTOR        : {
	expression << "sector(";
	break;
      }
      case ShapeBasics::BOX           : {
	expression << "box(";
	R[0] /= 2;
	R[1] /= 2;
	break;
      }
      case ShapeBasics::ROTBOX        : {
	expression << "box(";
	R[0] /= 2;
	R[1] /= 2;
	break;
      }
      case ShapeBasics::RECTANGLE     : {
	expression << "rectangle(";
	break;
      }
      case ShapeBasics::ROTRECTANGLE  : {
	expression << "rectangle(";
	break;
      }
      case ShapeBasics::RHOMBUS       : {
	expression << "diamond(";
	break;
      }
      case ShapeBasics::ROTRHOMBUS    : {
	expression << "diamond(";
	break;
      }
      case ShapeBasics::POLYGON       : {
	expression << "polygon(";
	break;
      }
      case ShapeBasics::LINE          : {
	expression << "line(";
	break;
      }
      }


      for(unsigned point=0; point<X.size(); point++){
	expression << X[point] << "," << Y[point];
	if( point != X.size()-1 ) expression << ",";
      }

      if( R.size() !=0 ) expression << ",";

      for(unsigned point=0; point<R.size(); point++){
	expression << R[point];
	if( point != R.size()-1 ) expression << ",";
      }

      if( angles.size() == 0 ){
	if( _region[comp][element]->type()==ShapeBasics::BOX ||
	    _region[comp][element]->type()==ShapeBasics::RECTANGLE ||
	    _region[comp][element]->type()==ShapeBasics::RHOMBUS )
	  expression << ",0";

      }else{
	expression << ",";
	
	for(unsigned point=0; point<angles.size(); point++){
	  expression << angles[point];
	  if( point != angles.size()-1 ) expression << ",";
	}
      }

      expression << "," << filtComps[0] << "," << filtComps[1] << ")";
      
      if( element != _region[comp].size()-1 ) expression << "&&";
    }

    expression << ")";
    if( comp != _region.size()-1 ) expression << "||";
  }
  expression << ")";

  return expression.str();
}


Range<double> DSregionFilter::extent(const std::string &axis) {
  int Xmin, Xmax, Ymin, Ymax;
  geoExtent(_region, Xmin, Xmax, Ymin, Ymax);

  DSfilterComp lcomps = comps();
  if( axis==lcomps[0] ){
    Range<double> r(Xmin, Xmax);
    return r;
  }else if( axis==lcomps[1] ){
    Range<double> r(Ymin, Ymax);
    return r;
  }

  errstr << "Attempt to extract extent information from the wrong filter"
	 << Sas::error("DSextentRegion");

  Range<double> r(0, 0, Unbounded, Unbounded);
  return r;
}

Range<double> DSregionFilter::realextent(const std::string &axis) {
  double Xmin, Xmax, Ymin, Ymax;
  realgeoExtent(_region, Xmin, Xmax, Ymin, Ymax);

  DSfilterComp lcomps = comps();
  if( axis==lcomps[0] ){
    Range<double> r(Xmin, Xmax);
    return r;
  }else if( axis==lcomps[1] ){
    Range<double> r(Ymin, Ymax);
    return r;
  }

  errstr << "Attempt to extract extent information from the wrong filter"
	 << Sas::error("DSextentRegion");

  Range<double> r(0, 0, Unbounded, Unbounded);
  return r;
}

double DSregionFilter::volume() {
  return geoArea(_region);
}


bool DSregionFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||


      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( region() != oFilter.region() ) return false;

  return true;
}


void DSregionFilter::operator!(){
  _region = _negateRegion(_region);
}


void geoExtent(const GeoRegion &region, int &Xmin, int &Xmax, int &Ymin, int &Ymax){

  for (unsigned comp=0; comp<region.size(); ++comp){
    for (unsigned element=0; element<region[comp].size(); ++element) {
      std::vector<float> X = region[comp][element]->pointsX();
      std::vector<float> Y = region[comp][element]->pointsY();
      std::vector<float> R = region[comp][element]->radii();

      if(region[comp][element]->type()==ShapeBasics::BOX ||
	 region[comp][element]->type()==ShapeBasics::ROTBOX){
	R[0] /=2;
	R[1] /=2;
      }

      float Rmax = 0;
      for(unsigned ir=0; ir<R.size(); ir++) 
	Rmax += R[ir]*R[ir];
      if(Rmax==0)
	for(unsigned i=1; i<X.size(); i++){
	  float iR = sqrt((X[0]-X[i])*(X[0]-X[i])+(Y[0]-Y[i])*(Y[0]-Y[i]));
	  if( iR>Rmax ) Rmax = iR;
	}
      else
	Rmax = sqrt(Rmax);

      if(comp==0 && element==0){
	Xmin = (int)floor(X[0]-Rmax);
	Xmax = (int)ceil(X[0]+Rmax);
	Ymin = (int)floor(Y[0]-Rmax);
	Ymax = (int)ceil(Y[0]+Rmax);
      }else{
	if( floor(X[0]-Rmax) < Xmin ) Xmin = (int)floor(X[0]-Rmax);
	if( ceil(X[0]+Rmax) > Xmax )  Xmax = (int)ceil(X[0]+Rmax);
	if( floor(Y[0]-Rmax) < Ymin ) Ymin = (int)floor(Y[0]-Rmax);
	if( ceil(Y[0]+Rmax) > Ymax )  Ymax = (int)ceil(Y[0]+Rmax);
      }
    }
  }

}

void realgeoExtent(const GeoRegion &region, double &Xmin, double &Xmax, double &Ymin, double &Ymax){

  for (unsigned comp=0; comp<region.size(); ++comp){
    for (unsigned element=0; element<region[comp].size(); ++element) {
      std::vector<float> X = region[comp][element]->pointsX();
      std::vector<float> Y = region[comp][element]->pointsY();
      std::vector<float> R = region[comp][element]->radii();

      if(region[comp][element]->type()==ShapeBasics::BOX ||
     region[comp][element]->type()==ShapeBasics::ROTBOX){
    R[0] /=2;
    R[1] /=2;
      }

      float Rmax = 0;
      for(unsigned ir=0; ir<R.size(); ir++)
    Rmax += R[ir]*R[ir];
      if(Rmax==0)
    for(unsigned i=1; i<X.size(); i++){
      float iR = sqrt((X[0]-X[i])*(X[0]-X[i])+(Y[0]-Y[i])*(Y[0]-Y[i]));
      if( iR>Rmax ) Rmax = iR;
    }
      else
    Rmax = sqrt(Rmax);

      if(comp==0 && element==0){
    Xmin = X[0]-Rmax;
    Xmax = X[0]+Rmax;
    Ymin = Y[0]-Rmax;
    Ymax = Y[0]+Rmax;
      }else{
    if( X[0]-Rmax < Xmin ) Xmin = X[0]-Rmax;
    if( X[0]+Rmax > Xmax )  Xmax =X[0]+Rmax;
    if( Y[0]-Rmax < Ymin ) Ymin = Y[0]-Rmax;
    if( Y[0]+Rmax > Ymax )  Ymax =Y[0]+Rmax;
      }
    }
  }

}

unsigned long geoArea(const GeoRegion &region){
  int Xmin, Xmax, Ymin, Ymax;

  geoExtent(region, Xmin, Xmax, Ymin, Ymax);

  DataSet *tempSet = dataSetServer->open(tempnam(0,0), Dal::Temp);
  Table *tempTab = tempSet->addTable("GRID", (Xmax-Xmin+1)*(Ymax-Ymin+1));
  ColumnData *Xdata = tempTab->addColumn("X", Dal::Int16)->data();
  ColumnData *Ydata = tempTab->addColumn("Y", Dal::Int16)->data();
  
  int16 *X = Xdata->int16Data();
  int16 *Y = Ydata->int16Data();

  unsigned long count = 0;
  for(int ix=Xmin; ix<=Xmax; ix++){
    for(int iy=Ymin; iy<=Ymax; iy++){
      X[count] = ix;
      Y[count] = iy;
      count++;
    }
  }

  delete Xdata;
  delete Ydata;

  DSregionFilter rFilter(NULL, "POS", region);
  DataFilter filter(rFilter.selectExpression());
  std::vector<DataFilter::RowRange> rowranges;

  unsigned long area = filter.doSelect(*tempTab, rowranges);
  dataSetServer->close(tempSet);

  return area;
}
