//
//  Implementation of the DSregionBLock class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <cstdio>
#include <sstream>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <ASCregion.h>
#include <DataSubspace.h>
#include <errstr.h>


DSregionBlock::DSregionBlock(DSregionBlock &dsblock) : DSblockFilter(dsblock) {}


DSregionBlock::DSregionBlock(DScomp *parent,
			     const std::string &name, const std::string &blockSpec,
			     DSfilterComp &filtercomp) :
  DSblockFilter(parent, RegionBlockT, name, blockSpec, filtercomp) {

    _openBlock();
    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "ASC" &&
	_block->attribute("HDUCLAS1")->asString() == "REGION" )
      return;
  
    errstr << "Block type is not compatible with the region filter type"
	   << Sas::error("BlockTypeRegion");
}


DSregionBlock::DSregionBlock(DScomp *parent,
			     const std::string &name, const std::string &blockSpec) :
  DSblockFilter(parent, RegionBlockT, name, blockSpec) {

    // Check that the block type is compatible

    _openBlock();
    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "ASC" &&
	_block->attribute("HDUCLAS1")->asString() == "REGION" )
      return;
  
    errstr << "Block type is not compatible with the region filter type"
	   << Sas::error("BlockTypeRegion");
}


std::string DSregionBlock::selectExpression() {
  std::stringstream expression;
  DSfilterComp lcomps = comps();

  expression << "region(" << _parentName << ":" << _blockName << "," 
	     <<  lcomps[0] << "," << lcomps[1] << ")";
  return expression.str();
}


Range<double> DSregionBlock::extent(const std::string &axis) {
  int Xmin, Xmax, Ymin, Ymax;

  InASCregionTable ascTab(blockSpec());
  GeoRegion region;

  ascTab >> region;
  geoExtent(region, Xmin, Xmax, Ymin, Ymax);

  DSfilterComp lcomps = comps();
  if( axis==lcomps[0] ){
    Range<double> r(Xmin, Xmax);
    return r;
  }else if( axis==lcomps[1] ){
    Range<double> r(Ymin, Ymax);
    return r;
  }

  errstr << "Attempt to extract extent information from the wrong filter"
	 << Sas::error("DSextentRegion");

  Range<double> r(0, 0, Unbounded, Unbounded);
  return r;
}

Range<double> DSregionBlock::realextent(const std::string &axis) {
  double Xmin, Xmax, Ymin, Ymax;

  InASCregionTable ascTab(blockSpec());
  GeoRegion region;

  ascTab >> region;
  realgeoExtent(region, Xmin, Xmax, Ymin, Ymax);

  DSfilterComp lcomps = comps();
  if( axis==lcomps[0] ){
    Range<double> r(Xmin, Xmax);
    return r;
  }else if( axis==lcomps[1] ){
    Range<double> r(Ymin, Ymax);
    return r;
  }

  errstr << "Attempt to extract extent information from the wrong filter"
	 << Sas::error("DSextentRegion");

  Range<double> r(0, 0, Unbounded, Unbounded);
  return r;
}


double DSregionBlock::volume() {
  return geoArea(region());
}


GeoRegion DSregionBlock::region() const{
  InASCregionTable ascTab(blockSpec());
  GeoRegion region;
  
  ascTab >> region;
  return region;
}


void DSregionBlock::operator!() {
  GeoRegion oldRegion = region();
  GeoRegion newRegion = _negateRegion(oldRegion);

  DataSet * tempset = dataSetServer->open(tempnam(0,0), DataSet::Temp);
  OutASCregionTable newTab(tempset);
  newTab << newRegion;

  DSfilterComp oldComps = comps();
  DSregionBlock newFilter(parent(), name(), tempset->name() + ":REGION", oldComps);
  *this = newFilter;
}
