!
!  Functions for utilizing the more basic F90 interface
!
!	AUTHOR:
!	  D. Hinshaw (DH)
!

module DSSutil

  use DSSinter
  use Dal

  implicit none

  private

  type, public :: BoundT
    integer             :: type
    real (kind=double)  :: value
  end type BoundT

  type, public :: RangeT
    type(BoundT)        :: lower, upper
  end type RangeT

  integer, public, parameter :: EXCLUSIVE = 1
  integer, public, parameter :: INCLUSIVE = 2
  integer, public, parameter :: UNDEFINED = 3

  public :: dssFilterRangesSize
  public :: dssComponentRangesSize
  public :: dssSubspaceRangesSize
  
  public dssRanges

  interface dssRanges
     module procedure dssSubspaceRanges
     module procedure dssComponentRanges
     module procedure dssFilterRanges
  end interface

contains

  function dssSubspaceRanges( dss, axis )

    type(RangeT), dimension(:), allocatable :: dssSubspaceRanges
    type(DssT), intent(in)              :: dss
    character(len=*), intent(in)        :: axis

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper
    integer                                  :: nSize, i

    call getDssRanges(dss, axis, lowerB, lower, upper, upperB)

    nSize = size(upper)
    if (allocated(dssSubspaceRanges)) deallocate( dssSubspaceRanges )  
    
    allocate( dssSubspaceRanges(0:nSize-1) )

    do i = 0, nSize-1
       dssSubspaceRanges(i)%lower%type = lowerB(i)
       dssSubspaceRanges(i)%upper%type = upperB(i)
       dssSubspaceRanges(i)%lower%value = lower(i)
       dssSubspaceRanges(i)%upper%value = upper(i)
    enddo

    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end function dssSubspaceRanges

  subroutine dssSubspaceRangesSize( dss, axis, nSize )

    integer, intent(out) :: nSize
    type(DssT), intent(in)              :: dss
    character(len=*), intent(in)        :: axis

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper

    call getDssRanges(dss, axis, lowerB, lower, upper, upperB)

    nSize = size(upper)

    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end subroutine dssSubspaceRangesSize



  function dssComponentRanges( dscomp, axis )

    type(RangeT), dimension(:), allocatable :: dssComponentRanges
    type(DScompT), intent(in)           :: dscomp
    character(len=*), intent(in)        :: axis

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper
    integer                                  :: nSize, i

    call getDssRanges(dscomp, axis, lowerB, lower, upper, upperB)

    nSize = size(upper)
    if (allocated(dssComponentRanges)) deallocate( dssComponentRanges )  
   
    allocate( dssComponentRanges(0:nSize-1) )

    do i = 0, nSize-1
       dssComponentRanges(i)%lower%type = lowerB(i)
       dssComponentRanges(i)%upper%type = upperB(i)
       dssComponentRanges(i)%lower%value = lower(i)
       dssComponentRanges(i)%upper%value = upper(i)
    enddo

    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end function dssComponentRanges

  subroutine dssComponentRangesSize( dscomp, axis, nSize )
	 
	integer, intent(out) :: nSize
    type(DScompT), intent(in)           :: dscomp
    character(len=*), intent(in)        :: axis

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper    

    call getDssRanges(dscomp, axis, lowerB, lower, upper, upperB)

    nSize = size(upper)

    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end subroutine dssComponentRangesSize



  function dssFilterRanges( dsfilter )

    type(RangeT), allocatable, dimension(:) :: dssFilterRanges
    type(DSfilterT), intent(in)         :: dsfilter

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper
    integer                                  :: nSize, i
    
    call getDssRanges(dsfilter, lowerB, lower, upper, upperB)

    nSize = size(upper)
    if (allocated(dssFilterRanges)) deallocate( dssFilterRanges )         
 	
 	allocate( dssFilterRanges(0:nSize-1) )
    
    do i = 0, nSize-1    
       dssFilterRanges(i)%lower%type = lowerB(i)
       dssFilterRanges(i)%upper%type = upperB(i)
       dssFilterRanges(i)%lower%value = lower(i)
       dssFilterRanges(i)%upper%value = upper(i)
    enddo

    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end function dssFilterRanges
  
  
  subroutine dssFilterRangesSize( dsfilter, nSize ) 

    integer, intent(out) :: nSize
    type(DSfilterT), intent(in)         :: dsfilter

    integer, dimension(:), pointer           :: lowerB, upperB
    real(kind=double), dimension(:), pointer :: lower, upper

    call getDssRanges(dsfilter, lowerB, lower, upper, upperB)

    nSize = size(upper)
    call releaseDssRanges(lowerB, lower, upper, upperB)
    
  end subroutine dssFilterRangesSize
  
end module DSSutil
